"""
This module provides REST API handlers for some requests concerning scenarios.
"""
import warnings
import os
import falcon

from ..base_handler import BaseHandler
from ..res_manager import ResourceManager
from ...utils import get_temp_folder, pack_zip_archive
from ...simulation import ScenarioSimulator, SensorConfig


class ScenarioManager(ResourceManager):
    """
    Class for managing all scenarios that are currently used by the REST API.
    """
    def create(self, **kwds) -> str:
        """
        Creates a new scenario -- e.g. loading a given .inp file or
        using a given scenario configuration.

        Returns
        -------
        `str`
            UUID of the new scenario.
        """
        return self.create_new_item(ScenarioSimulator(**kwds))

    def close_item(self, item: ScenarioSimulator) -> None:
        item.close()


class ScenarioBaseHandler(BaseHandler):
    """
    Base class for all handlers concerning scenarios.

    Parameters
    ----------
    scenario_mgr : :class:`~epyt_flow.rest_api.scenario.handlers.ScenarioManager`
        Instance for managing all scenarios.
    """
    def __init__(self, scenario_mgr: ScenarioManager):
        self.scenario_mgr = scenario_mgr


class ScenarioRemoveHandler(ScenarioBaseHandler):
    """
    Class for handling a DELETE request for a given scenario.
    """
    def on_delete(self, _, resp: falcon.Response, scenario_id: str) -> None:
        """
        Deletes a given scenario.

        Parameters
        ----------
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        scenario_id : `str`
            UUID of the scenario.
        """
        try:
            if self.scenario_mgr.validate_uuid(scenario_id) is False:
                self.send_invalid_resource_id_error(resp)
                return

            self.scenario_mgr.remove(scenario_id)
        except Exception as ex:
            warnings.warn(str(ex))
            resp.status = falcon.HTTP_INTERNAL_SERVER_ERROR


class ScenarioExportHandler(ScenarioBaseHandler):
    """
    Class for handling GET requests for exporting a given scenario to EPANET files
    -- i.e. .inp and (otpionally) .msx files.
    """
    def __create_temp_file_path(self, scenario_id: str, file_ext: str) -> None:
        """
        Returns a path to a temporary file for storing the scenario.

        Parameters
        ----------
        scenario_id : `str`
            UUID of the scenario.
        file_ext : `str`
            File extension.
        """
        return os.path.join(get_temp_folder(), f"{scenario_id}.{file_ext}")

    def __send_temp_file(self, resp: falcon.Response, tmp_file: str,
                         content_type: str = "application/octet-stream") -> None:
        """
        Sends a given file (`tmp_file`) to the the client.

        Parameters
        ----------
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        tmp_file : `str`
            Path to the temporary file to be send.
        """
        resp.status = falcon.HTTP_200
        resp.content_type = content_type
        with open(tmp_file, 'rb') as f:
            resp.text = f.read()

    def on_get(self, _, resp: falcon.Response, scenario_id: str) -> None:
        """
        Exports a given scenario to an .inp and (optionally) .msx file.

        Parameters
        ----------
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        scenario_id : `str`
            UUID of the scenario.
        """
        try:
            if self.scenario_mgr.validate_uuid(scenario_id) is False:
                self.send_invalid_resource_id_error(resp)
                return

            my_scenario = self.scenario_mgr.get(scenario_id)

            f_inp_out = self.__create_temp_file_path(scenario_id, "inp")
            f_msx_out = self.__create_temp_file_path(scenario_id, "msx")
            my_scenario.save_to_epanet_file(f_inp_out, f_msx_out)

            if os.path.isfile(f_msx_out):
                f_out = self.__create_temp_file_path(scenario_id, "zip")
                pack_zip_archive([f_inp_out, f_msx_out], f_out)

                self.__send_temp_file(resp, f_out)
                os.remove(f_out)
                os.remove(f_msx_out)
            else:
                self.__send_temp_file(resp, f_inp_out)
            os.remove(f_inp_out)
        except Exception as ex:
            warnings.warn(str(ex))
            resp.status = falcon.HTTP_INTERNAL_SERVER_ERROR


class ScenarioConfigHandler(ScenarioBaseHandler):
    """
    Class for handling a GET request for getting the scenario configuration of a given scenario.
    """
    def on_get(self, _, resp: falcon.Response, scenario_id: str) -> None:
        """
        Gets the scenario configuration of a given scenario.

        Parameters
        ----------
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        scenario_id : `str`
            UUID of the scenario.
        """
        try:
            if self.scenario_mgr.validate_uuid(scenario_id) is False:
                self.send_invalid_resource_id_error(resp)
                return

            my_sceanrio_config = self.scenario_mgr.get(scenario_id).get_scenario_config()
            self.send_json_response(resp, my_sceanrio_config)
        except Exception as ex:
            warnings.warn(str(ex))
            resp.status = falcon.HTTP_INTERNAL_SERVER_ERROR


class ScenarioNewHandler(ScenarioBaseHandler):
    """
    Class for handling POST requests for creating a new scenario.
    """
    def on_post(self, req: falcon.Request, resp: falcon.Response) -> None:
        """
        Creates/Loads a new scenario.

        Parameters
        ----------
        req : `falcon.Request <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#request>`_
            Request instance.
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        scenario_id : `str`
            UUID of the scenario.
        """
        try:
            args = self.load_json_data_from_request(req)
            scenario_id = self.scenario_mgr.create(**args)
            self.send_json_response(resp, {"scenario_id": scenario_id})
        except Exception as ex:
            warnings.warn(str(ex))
            resp.status = falcon.HTTP_INTERNAL_SERVER_ERROR


class ScenarioTopologyHandler(ScenarioBaseHandler):
    """
    Class for handling GET requests for getting the topology of a given scenario.
    """
    def on_get(self, _, resp: falcon.Response, scenario_id: str) -> None:
        """
        Gets the topology of a given scenario.

        Parameters
        ----------
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        scenario_id : `str`
            UUID of the scenario.
        """
        try:
            if self.scenario_mgr.validate_uuid(scenario_id) is False:
                self.send_invalid_resource_id_error(resp)
                return

            my_topology = self.scenario_mgr.get(scenario_id).get_topology()
            self.send_json_response(resp, my_topology)
        except Exception as ex:
            warnings.warn(str(ex))
            resp.status = falcon.HTTP_INTERNAL_SERVER_ERROR


class ScenarioGeneralParamsHandler(ScenarioBaseHandler):
    """
    Class for handling GET and POST requests for the general parameters of a given scenario.
    """
    def on_get(self, _, resp: falcon.Response, scenario_id: str) -> None:
        """
        Gets the general parameters (e.g. simulation duration, etc.) of a given scenario.

        Parameters
        ----------
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        scenario_id : `str`
            UUID of the scenario.
        """
        try:
            if self.scenario_mgr.validate_uuid(scenario_id) is False:
                self.send_invalid_resource_id_error(resp)
                return

            my_general_params = self.scenario_mgr.get(scenario_id).get_scenario_config().\
                general_params
            self.send_json_response(resp, my_general_params)
        except Exception as ex:
            warnings.warn(str(ex))
            resp.status = falcon.HTTP_INTERNAL_SERVER_ERROR

    def on_post(self, req: falcon.Request, resp: falcon.Response, scenario_id: str) -> None:
        """
        Sets the general parameters of a given scenario.

        Parameters
        ----------
        req : `falcon.Request <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#request>`_
            Request instance.
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Request instance.
        scenario_id : `str`
            UUID of the scenario.
        """
        try:
            if self.scenario_mgr.validate_uuid(scenario_id) is False:
                self.send_invalid_resource_id_error(resp)
                return

            general_params = self.load_json_data_from_request(req)
            if not isinstance(general_params, dict):
                self.send_json_parsing_error(resp)
                return

            self.scenario_mgr.get(scenario_id).set_general_parameters(**general_params)
        except Exception as ex:
            warnings.warn(str(ex))
            resp.status = falcon.HTTP_INTERNAL_SERVER_ERROR


class ScenarioQualityParamsHandler(ScenarioBaseHandler):
    """
    Class for handling POST requests for specifying (EPANET) quality parameters of a given scenario.
    """
    def on_post(self, req: falcon.Request, resp: falcon.Response, scenario_id: str) -> None:
        """
        Specifies the (EPANET) quality parameters of a given scenario.

        Parameters
        ----------
        req : `falcon.Request <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#request>`_
            Request instance.
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Request instance.
        scenario_id : `str`
            UUID of the scenario.
        """
        try:
            if self.scenario_mgr.validate_uuid(scenario_id) is False:
                self.send_invalid_resource_id_error(resp)
                return

            quality_params = self.load_json_data_from_request(req)
            if not isinstance(quality_params, dict):
                self.send_json_parsing_error(resp)
                return

            self.scenario_mgr.get(scenario_id).set_quality_parameters(**quality_params)
        except Exception as ex:
            warnings.warn(str(ex))
            resp.status = falcon.HTTP_INTERNAL_SERVER_ERROR


class ScenarioSensorConfigHandler(ScenarioBaseHandler):
    """
    Class for handling GET and POST requests for the sensor configuration of a given scenario.
    """
    def on_get(self, _, resp: falcon.Response, scenario_id: str) -> None:
        """
        Gets the sensor configuration of a given scenario.

        Parameters
        ----------
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        scenario_id : `str`
            UUID of the scenario.
        """
        try:
            if self.scenario_mgr.validate_uuid(scenario_id) is False:
                self.send_invalid_resource_id_error(resp)
                return

            my_sensor_config = self.scenario_mgr.get(scenario_id).sensor_config
            self.send_json_response(resp, my_sensor_config)
        except Exception as ex:
            warnings.warn(str(ex))
            resp.status = falcon.HTTP_INTERNAL_SERVER_ERROR

    def on_post(self, req: falcon.Request, resp: falcon.Response, scenario_id: str) -> None:
        """
        Sets the sensor configuration of a given scenario.

        Parameters
        ----------
        req : `falcon.Request <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#request>`_
            Request instance.
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Request instance.
        scenario_id : `str`
            UUID of the scenario.
        """
        try:
            if self.scenario_mgr.validate_uuid(scenario_id) is False:
                self.send_invalid_resource_id_error(resp)
                return

            sensor_config = self.load_json_data_from_request(req)
            if not isinstance(sensor_config, SensorConfig):
                self.send_json_parsing_error(resp)
                return

            my_scenario = self.scenario_mgr.get(scenario_id)
            my_scenario.sensor_config = sensor_config
        except Exception as ex:
            warnings.warn(str(ex))
            resp.status = falcon.HTTP_INTERNAL_SERVER_ERROR


class ScenarioNodeDemandPatternHandler(ScenarioBaseHandler):
    """
    Class for handling POST requests for node demand patterns of a given scenario.
    """
    def on_post(self, req: falcon.Request, resp: falcon.Response, scenario_id: str,
                node_id: str) -> None:
        """
        Sets the demand pattern of a specific node in a given scenario.

        Parameters
        ----------
        req : `falcon.Request <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#request>`_
            Request instance.
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        scenario_id : `str`
            UUID of the scenario.
        node_id : `str`
            ID of the node.
        """
        try:
            if self.scenario_mgr.validate_uuid(scenario_id) is False:
                self.send_invalid_resource_id_error(resp)
                return

            params = self.load_json_data_from_request(req)

            my_scenario = self.scenario_mgr.get(scenario_id)
            my_scenario.set_node_demand_pattern(node_id, params["base_demand"],
                                                params["demand_pattern_id"],
                                                params["demand_pattern"])
        except Exception as ex:
            warnings.warn(str(ex))
            resp.data = str(ex)
            resp.status = falcon.HTTP_INTERNAL_SERVER_ERROR



"""
This module provides REST API handlers for accessing the final sensor readings
(e.g. pressure, flow rate, etc.).
"""
import warnings
import falcon

from .handlers import ScadaDataBaseHandler


class ScadaDataPressuresHandler(ScadaDataBaseHandler):
    """
    Class for handling GET requests for the pressure sensor readings of a given SCADA data instance.
    """
    def on_get(self, _, resp: falcon.Response, data_id: str) -> None:
        """
        Gets the pressure sensor readings of a given SCADA data instance.

        Parameters
        ----------
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        data_id : `str`
            UUID of the SCADA data.
        """
        try:
            if self.scada_data_mgr.validate_uuid(data_id) is False:
                self.send_invalid_resource_id_error(resp)
                return

            data_pressures = self.scada_data_mgr.get(data_id).get_data_pressures().tolist()
            self.send_json_response(resp, data_pressures)
        except Exception as ex:
            warnings.warn(str(ex))
            resp.status = falcon.HTTP_INTERNAL_SERVER_ERROR


class ScadaDataFlowsHandler(ScadaDataBaseHandler):
    """
    Class for handling GET requests for the flow sensor readings of a given SCADA data instance.
    """
    def on_get(self, _, resp: falcon.Response, data_id: str) -> None:
        """
        Gets the flow sensor readings of a given SCADA data instance.

        Parameters
        ----------
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        data_id : `str`
            UUID of the SCADA data.
        """
        try:
            if self.scada_data_mgr.validate_uuid(data_id) is False:
                self.send_invalid_resource_id_error(resp)
                return

            data_flows = self.scada_data_mgr.get(data_id).get_data_flows().tolist()
            self.send_json_response(resp, data_flows)
        except Exception as ex:
            warnings.warn(str(ex))
            resp.status = falcon.HTTP_INTERNAL_SERVER_ERROR


class ScadaDataDemandsHandler(ScadaDataBaseHandler):
    """
    Class for handling GET requests for the demand sensor readings of a given SCADA data instance.
    """
    def on_get(self, _, resp: falcon.Response, data_id: str) -> None:
        """
        Gets the demand sensor readings of a given SCADA data instance.

        Parameters
        ----------
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        data_id : `str`
            UUID of the SCADA data.
        """
        try:
            if self.scada_data_mgr.validate_uuid(data_id) is False:
                self.send_invalid_resource_id_error(resp)
                return

            data_demands = self.scada_data_mgr.get(data_id).get_data_demands().tolist()
            self.send_json_response(resp, data_demands)
        except Exception as ex:
            warnings.warn(str(ex))
            resp.status = falcon.HTTP_INTERNAL_SERVER_ERROR


class ScadaDataValveStatesHandler(ScadaDataBaseHandler):
    """
    Class for handling GET requests for the valve state sensor readings
    of a given SCADA data instance.
    """
    def on_get(self, _, resp: falcon.Response, data_id: str) -> None:
        """
        Gets the valve state sensor readings of a given SCADA data instance.

        Parameters
        ----------
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        data_id : `str`
            UUID of the SCADA data.
        """
        try:
            if self.scada_data_mgr.validate_uuid(data_id) is False:
                self.send_invalid_resource_id_error(resp)
                return

            data_demands = self.scada_data_mgr.get(data_id).get_data_valves_state().tolist()
            self.send_json_response(resp, data_demands)
        except Exception as ex:
            warnings.warn(str(ex))
            resp.status = falcon.HTTP_INTERNAL_SERVER_ERROR


class ScadaDataPumpStatesHandler(ScadaDataBaseHandler):
    """
    Class for handling GET requests for the pump state sensor readings
    of a given SCADA data instance.
    """
    def on_get(self, _, resp: falcon.Response, data_id: str) -> None:
        """
        Gets the pump state sensor readings of a given SCADA data instance.

        Parameters
        ----------
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        data_id : `str`
            UUID of the SCADA data.
        """
        try:
            if self.scada_data_mgr.validate_uuid(data_id) is False:
                self.send_invalid_resource_id_error(resp)
                return

            data_demands = self.scada_data_mgr.get(data_id).get_data_pumps_state().tolist()
            self.send_json_response(resp, data_demands)
        except Exception as ex:
            warnings.warn(str(ex))
            resp.status = falcon.HTTP_INTERNAL_SERVER_ERROR


class ScadaDataTankVolumesHandler(ScadaDataBaseHandler):
    """
    Class for handling GET requests for the tank volume sensor readings
    of a given SCADA data instance.
    """
    def on_get(self, _, resp: falcon.Response, data_id: str) -> None:
        """
        Gets the tank volume sensor readings of a given SCADA data instance.

        Parameters
        ----------
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        data_id : `str`
            UUID of the SCADA data.
        """
        try:
            if self.scada_data_mgr.validate_uuid(data_id) is False:
                self.send_invalid_resource_id_error(resp)
                return

            data_demands = self.scada_data_mgr.get(data_id).get_data_tanks_water_volume().tolist()
            self.send_json_response(resp, data_demands)
        except Exception as ex:
            warnings.warn(str(ex))
            resp.status = falcon.HTTP_INTERNAL_SERVER_ERROR


class ScadaDataNodesQualityHandler(ScadaDataBaseHandler):
    """
    Class for handling GET requests for the node quality sensor readings
    of a given SCADA data instance.
    """
    def on_get(self, _, resp: falcon.Response, data_id: str) -> None:
        """
        Gets the node quality sensor readings of a given SCADA data instance.

        Parameters
        ----------
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        data_id : `str`
            UUID of the SCADA data.
        """
        try:
            if self.scada_data_mgr.validate_uuid(data_id) is False:
                self.send_invalid_resource_id_error(resp)
                return

            data_nodes_quality = self.scada_data_mgr.get(data_id).get_data_nodes_quality().tolist()
            self.send_json_response(resp, data_nodes_quality)
        except Exception as ex:
            warnings.warn(str(ex))
            resp.status = falcon.HTTP_INTERNAL_SERVER_ERROR


class ScadaDataLinksQualityHandler(ScadaDataBaseHandler):
    """
    Class for handling GET requests for the link quality sensor readings
    of a given SCADA data instance.
    """
    def on_get(self, _, resp: falcon.Response, data_id: str) -> None:
        """
        Gets the link/pipe quality sensor readings of a given SCADA data instance.

        Parameters
        ----------
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        data_id : `str`
            UUID of the SCADA data.
        """
        try:
            if self.scada_data_mgr.validate_uuid(data_id) is False:
                self.send_invalid_resource_id_error(resp)
                return

            data_links_quality = self.scada_data_mgr.get(data_id).get_data_links_quality().tolist()
            self.send_json_response(resp, data_links_quality)
        except Exception as ex:
            warnings.warn(str(ex))
            resp.status = falcon.HTTP_INTERNAL_SERVER_ERROR


class ScadaDataNodeBulkSpeciesHandler(ScadaDataBaseHandler):
    """
    Class for handling GET requests for the bulk species node concentration sensor readings
    of a given SCADA data instance.
    """
    def on_get(self, _, resp: falcon.Response, data_id: str) -> None:
        """
        Gets the bulk species node concentrations sensor readings of a given SCADA data instance.

        Parameters
        ----------
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        data_id : `str`
            UUID of the SCADA data.
        """
        try:
            if self.scada_data_mgr.validate_uuid(data_id) is False:
                self.send_invalid_resource_id_error(resp)
                return

            data_conc = self.scada_data_mgr.get(data_id).\
                get_data_bulk_species_node_concentration().tolist()
            self.send_json_response(resp, data_conc)
        except Exception as ex:
            warnings.warn(str(ex))
            resp.status = falcon.HTTP_INTERNAL_SERVER_ERROR


class ScadaDataLinkBulkSpeciesHandler(ScadaDataBaseHandler):
    """
    Class for handling GET requests for the bulk species link/pipe concentration sensor readings
    of a given SCADA data instance.
    """
    def on_get(self, _, resp: falcon.Response, data_id: str) -> None:
        """
        Gets the bulk species link/pipe concentrations sensor readings of a
        given SCADA data instance.

        Parameters
        ----------
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        data_id : `str`
            UUID of the SCADA data.
        """
        try:
            if self.scada_data_mgr.validate_uuid(data_id) is False:
                self.send_invalid_resource_id_error(resp)
                return

            data_conc = self.scada_data_mgr.get(data_id).\
                get_data_bulk_species_link_concentration().tolist()
            self.send_json_response(resp, data_conc)
        except Exception as ex:
            warnings.warn(str(ex))
            resp.status = falcon.HTTP_INTERNAL_SERVER_ERROR


class ScadaDataSurfaceSpeciesHandler(ScadaDataBaseHandler):
    """
    Class for handling GET requests for the surface species concentration sensor readings
    of a given SCADA data instance.
    """
    def on_get(self, _, resp: falcon.Response, data_id: str) -> None:
        """
        Gets the surface species concentrations sensor readings of a given SCADA data instance.

        Parameters
        ----------
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        data_id : `str`
            UUID of the SCADA data.
        """
        try:
            if self.scada_data_mgr.validate_uuid(data_id) is False:
                self.send_invalid_resource_id_error(resp)
                return

            data_links_quality = self.scada_data_mgr.get(data_id).\
                get_data_surface_species_concentration().tolist()
            self.send_json_response(resp, data_links_quality)
        except Exception as ex:
            warnings.warn(str(ex))
            resp.status = falcon.HTTP_INTERNAL_SERVER_ERROR



"""
Module provides functions for simulating several scenarios in parallel.
"""
from typing import Callable, Any
import os
import warnings
import shutil
from multiprocess import Pool, cpu_count
import psutil

from .scenario_config import ScenarioConfig
from .scada import ScadaData
from .scenario_simulator import ScenarioSimulator


def callback_save_to_file(folder_out: str = "") -> Callable[[ScadaData, ScenarioConfig, int], None]:
    """
    Creates a callback for storing the simulation results in a .epytflow_scada_data file.
    The returned callback can be directly passed to
    :func:`~epyt_flow.simulation.parallel_simulation.ParallelScenarioSimulation.run`.

    Parameters
    ----------
    folder_out : `str`, optional
        Path to the folder where the simulation results will be stored.

        The default is the current working directory.

    Returns
    -------
    `Callable[[ScadaData, ScenarioConfig, int], None]`
        Callback storing the simulation results.
    """
    def callback(scada_data: ScadaData, _, scenario_idx: int) -> None:
        scada_data.save_to_file(os.path.join(folder_out, f"{scenario_idx}"))

    return callback


def _run_scenario_simulation(scenario_config: ScenarioConfig, scenario_idx: int,
                             callback: Callable[[ScadaData, ScenarioConfig, int], Any]) -> Any:
    with ScenarioSimulator(scenario_config=scenario_config) as sim:
        return callback(sim.run_simulation(), scenario_config, scenario_idx)


class ParallelScenarioSimulation():
    """
    Class providing functions to run scenario simulations in parallel.
    """
    @staticmethod
    def run(scenarios: list[ScenarioConfig], n_jobs: int = -1,
            max_working_memory_consumption: int = None,
            callback: Callable[[ScadaData, ScenarioConfig, int], Any] = callback_save_to_file()
            ) -> Any:
        """
        Simulates multiple scenarios in parallel.

        Parameters
        ----------
        scenarios : list[:class:`~epyt_flow.simulation.scenario_config.ScenarioConfig`]
            List of scenarios to be simulated.
        n_jobs : `int`, optional
            Number of CPUs that can be used by the simulations -- usually, this translates to
            the number of scenarios that are simulated in parallel.

            If -1, all CPUs are used.

            The default is -1
        max_working_memory_consumption : `int`, optional
            Maximum amount of working memory in MB that can be used by the simulations.
            Note that this might limit the number of scenarios that can be simulated in parallel.

            The default is None.
        callback: `Callable[[ScadaData, ScenarioConfig, int], None]`, optional
            Callback that is called after the simulation of a scenario finished.

            The callback gets the simulation results as a
            :class:`~epyt_flow.simulation.scada.scada_data.ScadaData` instance, the scenario
            configuration as a :class:`~epyt_flow.simulation.scenario_config.ScenarioConfig`
            instance, and the index of the scenario in 'scenarios' as arguments.

            The default is :func:`~epyt_flow.simulation.parallel_simulation.callback_save_to_file`.
        """
        if not isinstance(scenarios, list):
            raise TypeError("'scenarios' must be an instance of 'list[ScenarioConfig]' " +
                            f"but not of '{type(scenarios)}'")
        if any(not isinstance(item, ScenarioConfig) for item in scenarios):
            raise TypeError("Each item in 'scenarios' must be an instance of 'ScenarioConfig'")

        if not isinstance(n_jobs, int):
            raise TypeError(f"'n_jobs' must be an instance of 'int' but not of '{type(n_jobs)}'")
        if not (n_jobs == -1 or n_jobs > 0):
            raise ValueError("'n_jobs' must be either -1 or a positive integer")

        if max_working_memory_consumption is not None:
            if not isinstance(max_working_memory_consumption, int) or \
                    max_working_memory_consumption <= 0:
                raise ValueError("'max_working_memory_consumption' must be a positive integer")

        if not callable(callback):
            raise TypeError("'callback' mut be a callable " +
                            "'Callable[[ScadaData, ScenarioConfig, int], None]'")

        # Get free memory in MB
        ram_free_memory = psutil.virtual_memory().free * .000001
        if max_working_memory_consumption is not None:
            ram_free_memory = max(ram_free_memory, max_working_memory_consumption)

        harddisk_free_memory = shutil.disk_usage(".").free * .000001

        # Check memory requirements of each scenario
        max_memory_required = max(s_config.memory_consumption_estimate
                                  if s_config.memory_consumption_estimate is not None else 0
                                  for s_config in scenarios)
        if max_memory_required > ram_free_memory:
            raise RuntimeError("Not enough working memory avaialble! " +
                               f"Requested {max_memory_required} MB but only " +
                               f"{ram_free_memory} MB are available")

        if sum(s_config.memory_consumption_estimate
               if s_config.memory_consumption_estimate is not None else 0
               for s_config in scenarios) >= harddisk_free_memory:
            warnings.warn("There might not be enough free space on the hard disk " +
                          "to store all scenario results")

        # Compute number of processes that can run in parallel
        n_available_cpus = cpu_count()
        if n_jobs != -1:
            n_available_cpus = min(n_available_cpus, n_jobs)

        required_memory_bound = min(ram_free_memory, harddisk_free_memory)
        n_max_parallel_scenarios = n_available_cpus
        if max_memory_required != 0:
            n_max_parallel_scenarios = int(required_memory_bound / max_memory_required)

        n_parallel_scenarios = min(n_available_cpus, n_max_parallel_scenarios)

        if any(s_config.f_msx_in is not None for s_config in scenarios):
            n_parallel_scenarios = 1

        # Run scenario simulations
        scenarios_task = []
        for scenario_idx, scenario in enumerate(scenarios):
            scenarios_task.append((scenario, scenario_idx, callback))

        with Pool(processes=n_parallel_scenarios, maxtasksperchild=1) as pool:
            return pool.starmap(_run_scenario_simulation, scenarios_task)



"""
Module provides a class for implementing species injection (e.g. contamination) events.
"""
from copy import deepcopy
import warnings
import math
import numpy as np
from epanet_plus import EPyT, EpanetConstants

from .system_event import SystemEvent
from ...serialization import serializable, JsonSerializable, \
    SPECIESINJECTION_EVENT_ID


@serializable(SPECIESINJECTION_EVENT_ID, ".epytflow_speciesinjection_event")
class SpeciesInjectionEvent(SystemEvent, JsonSerializable):
    """
    Class implementing a (bulk) species injection event -- e.g. modeling a contamination event.

    Parameters
    ----------
    species_id : `str`
        ID of the bulk species that is going to be injected.
    node_id : `str`
        ID of the node at which the injection is palced.
    profile : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_
        Injection strength profile -- i.e. every entry corresponds to the strength of the injection
        at a point in time. Pattern will repeat if it is shorter than the total injection time.

        Note that the pattern time step is equivalent to the EPANET pattern time step.
    source_type : `int`
        Type of the bulk species injection source -- must be one of
        the following EPANET constants:

            - EN_CONCEN     = 0
            - EN_MASS       = 1
            - EN_SETPOINT   = 2
            - EN_FLOWPACED  = 3

        Description:

            - E_CONCEN Sets the concentration of external inflow entering a node
            - EN_MASS Injects a given mass/minute into a node
            - EN_SETPOINT Sets the concentration leaving a node to a given value
            - EN_FLOWPACED Adds a given value to the concentration leaving a node
    """
    def __init__(self, species_id: str, node_id: str, profile: np.ndarray, source_type: int,
                 **kwds):
        if not isinstance(species_id, str):
            raise TypeError("'species_id' must be an instance of 'str' but not of " +
                            f"'{type(species_id)}'")
        if not isinstance(node_id, str):
            raise TypeError("'node_id' must be an instance of 'str' but not of " +
                            f"'{type(node_id)}'")
        if not isinstance(profile, np.ndarray):
            raise TypeError("'profile' must be an instance of 'numpy.ndarray' but not of " +
                            f"'{type(profile)}'")
        if not isinstance(source_type, int):
            raise TypeError("'source_type' must be an instance of 'int' but not of " +
                            f"'{type(source_type)}'")
        if not 0 <= source_type <= 3:
            raise ValueError("'source_tye' must be in [0, 3]")

        self.__species_id = species_id
        self.__node_id = node_id
        self._profile = profile
        self.__source_type = source_type

        super().__init__(**kwds)

    @property
    def species_id(self) -> str:
        """
        Gets the ID of the bulk species that is going to be injected.

        Returns
        -------
        `str`
            Bulk species ID.
        """
        return self.__species_id

    @property
    def node_id(self) -> str:
        """
        Gets the ID of the node at which the injection is palced.

        Returns
        -------
        `str`
            Node ID.
        """
        return self.__node_id

    @property
    def profile(self) -> np.ndarray:
        """
        Gets the injection strength profile.

        Returns
        -------
        `numpy.ndarray`
            Pattern of the injection.
        """
        return deepcopy(self._profile)

    @property
    def source_type(self) -> int:
        """
        Type of the bulk species injection source -- will be one of
        the following EPANET toolkit constants:

            - EN_CONCEN     = 0
            - EN_MASS       = 1
            - EN_SETPOINT   = 2
            - EN_FLOWPACED  = 3

        Returns
        -------
        `int`
            Type of the injection source.
        """
        return self.__source_type

    def get_attributes(self) -> dict:
        return super().get_attributes() | {"species_id": self.__species_id,
                                           "node_id": self.__node_id, "profile": self._profile,
                                           "source_type": self.__source_type}

    def __eq__(self, other) -> bool:
        return super().__eq__(other) and self.__species_id == other.species_id and \
            self.__node_id == other.node_id and np.all(self._profile == other.profile) and \
            self.__source_type == other.source_type

    def __str__(self) -> str:
        return f"{super().__str__()} species_id: {self.__species_id} " +\
            f"node_id: {self.__node_id} profile: {self._profile} source_type: {self.__source_type}"

    def _get_pattern_id(self) -> str:
        return f"{self.__species_id}_{self.__node_id}"

    def init(self, epanet_api: EPyT) -> None:
        super().init(epanet_api)

        # Check parameters
        if self.__species_id not in self._epanet_api.get_all_msx_species_id():
            raise ValueError(f"Unknown species '{self.__species_id}'")
        if self.__node_id not in self._epanet_api.get_all_nodes_id():
            raise ValueError(f"Unknown node '{self.__node_id}'")

        # Create final injection strength pattern
        total_sim_duration = self._epanet_api.get_simulation_duration()
        time_step = self._epanet_api.get_hydraulic_time_step()

        pattern = np.zeros(math.ceil(total_sim_duration / time_step))

        end_time = self.end_time if self.end_time is not None else total_sim_duration
        injection_pattern_length = math.ceil((end_time - self.start_time) / time_step)
        injection_time_start_idx = int(self.start_time / time_step)

        injection_pattern = None
        if len(self._profile) == injection_pattern_length:
            injection_pattern = self.profile
        else:
            injection_pattern = np.tile(self.profile,
                                        math.ceil(injection_pattern_length / len(self.profile)))

        pattern[injection_time_start_idx:
                injection_time_start_idx + injection_pattern_length] = injection_pattern

        # Create injection
        pattern_id = self._get_pattern_id()
        if pattern_id in [self._epanet_api.MSXgetID(EpanetConstants.MSX_PATTERN, pattern_idx + 1)
                          for pattern_idx in
                          range(self._epanet_api.MSXgetcount(EpanetConstants.MSX_PATTERN))]:
            node_idx = self._epanet_api.get_node_idx(self.__node_id)
            species_idx = self._epanet_api.get_msx_species_idx(self.__species_id)
            cur_source_type = self._epanet_api.MSXgetsource(node_idx, species_idx)
            if cur_source_type[0] != self.__source_type:
                raise ValueError("Source type does not match existing source type")

            # Add new injection amount to existing injection --
            # i.e. two injection events at the same node
            pattern_idx = self._epanet_api.MSXgetindex(EpanetConstants.MSX_PATTERN, pattern_id)
            cur_pattern = self._epanet_api.get_msx_pattern(pattern_idx)
            cur_pattern = np.array(cur_pattern) + pattern
            self._epanet_api.MSXsetpattern(pattern_idx, cur_pattern.tolist(), len(cur_pattern))
        else:
            self._epanet_api.add_msx_pattern(pattern_id, pattern.tolist())
        self._epanet_api.set_msx_source(self.__node_id, self.__species_id, self.__source_type, 1,
                                        pattern_id)

    def cleanup(self) -> None:
        warnings.warn("Can not undo SpeciesInjectionEvent -- " +
                      "EPANET-MSX does not support removing patterns")

    def apply(self, cur_time: int) -> None:
        pass



"""
Module provides classes for implementing different types of uncertainties.
"""
from abc import ABC, abstractmethod
from copy import deepcopy
import numpy as np
from numpy.random import Generator, default_rng

from .utils import generate_deep_random_gaussian_noise, create_deep_random_pattern
from ..serialization import serializable, JsonSerializable, ABSOLUTE_GAUSSIAN_UNCERTAINTY_ID, \
    RELATIVE_GAUSSIAN_UNCERTAINTY_ID, ABSOLUTE_UNIFORM_UNCERTAINTY_ID, \
    RELATIVE_UNIFORM_UNCERTAINTY_ID, ABSOLUTE_DEEP_UNIFORM_UNCERTAINTY_ID, \
    RELATIVE_DEEP_UNIFORM_UNCERTAINTY_ID, ABSOLUTE_DEEP_GAUSSIAN_UNCERTAINTY_ID, \
    RELATIVE_DEEP_GAUSSIAN_UNCERTAINTY_ID, ABSOLUTE_DEEP_UNCERTAINTY_ID, \
    RELATIVE_DEEP_UNCERTAINTY_ID, PERCENTAGE_DEVIATON_UNCERTAINTY_ID


class Uncertainty(ABC):
    """
    Base class for uncertainties -- i.e. perturbations of data/signals.

    Parameters
    ----------
    min_value : `float`, optional
        Lower bound on the data/signal that is perturbed by this uncertainty.

        The default is None.
    max_value : `float`, optional
        Upper bound on the data/signal that is perturbed by this uncertainty.

        The default is None.
    """
    def __init__(self, min_value: float = None, max_value: float = None, **kwds):
        super().__init__(**kwds)

        self.__min_value = min_value
        self.__max_value = max_value

        self._random_generator = default_rng()

    @property
    def min_value(self) -> float:
        """
        Gets the lower bound on the data/signal.

        Returns
        -------
        `float`
            Lower bound.
        """
        return self.__min_value

    @property
    def max_value(self) -> float:
        """
        Gets the upper bound on the data/signal.

        Returns
        -------
        `float`
            Upper bound.
        """
        return self.__max_value

    @property
    def random_generator(self) -> Generator:
        """
        Returns the random number generator that is used for generating the uncertainties.

        Returns
        -------
        `numpy.random.Generator <https://numpy.org/doc/stable/reference/random/generator.html#numpy.random.Generator>`_
            The random number generator.
        """
        return deepcopy(self._random_generator)

    def set_random_generator(self, np_rand_generator: Generator) -> None:
        """
        Sets the random number generator that is going to be used for generating the uncertainties.

        Parameters
        ----------
        np_rand_generator : `numpy.random.Generator <https://numpy.org/doc/stable/reference/random/generator.html#numpy.random.Generator>`_, optional
            The random number generator.

            The default is the default BitGenerator (PCG64) as constructed by
            `numpy.random.default_rng() <https://numpy.org/doc/stable/reference/random/generator.html#numpy.random.default_rng>`_.
        """
        self._random_generator = np_rand_generator

    def get_attributes(self) -> dict:
        """
        Gets all attributes to be serialized -- these attributes are passed to the
        constructor when the object is deserialized.

        Returns
        -------
        `dict`
            Dictionary of attributes -- i.e. pairs of attribute name + value.
        """
        return {"min_value": self.__min_value, "max_value": self.__max_value}

    def __eq__(self, other) -> bool:
        if not isinstance(other, Uncertainty):
            raise TypeError("Can not compare 'Uncertainty' instance " +
                            f"with '{type(other)}' instance")

        return self.__min_value == other.min_value and self.__max_value == other.max_value and \
            self._random_generator == other.random_generator

    def __str__(self) -> str:
        return f"min_value: {self.__min_value} max_value: {self.__max_value}"

    def clip(self, data: np.ndarray) -> np.ndarray:
        """
        Clips values in a given array -- i.e. every value must be in [min_value, max_value].

        Parameters
        ----------
        data : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_
            Array to be clipped.

        Returns
        -------
        `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_
            Clipped data.
        """
        if self.__min_value is not None:
            data = np.max([data, self.__min_value])
        if self.__max_value is not None:
            data = np.min([data, self.__max_value])

        return data

    @abstractmethod
    def apply(self, data: float):
        """
        Applies the uncertainty to a single value.

        Parameters
        ----------
        data : `float`
            The value to which the uncertainty is applied.

        Returns
        -------
        `float`
            Uncertainty applied to 'data'.
        """
        raise NotImplementedError()

    def apply_batch(self, data: np.ndarray) -> np.ndarray:
        """
        Applies the uncertainty to an array of values.

        Parameters
        ----------
        data : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_
            Array of values to which the uncertainty is applied.

        Returns
        -------
        `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_
            Uncertainty applied to `data`.
        """
        for t in range(data.shape[0]):
            data[t] = self.apply(data[t])
        return data


class GaussianUncertainty(Uncertainty):
    """
    Base class implementing Gaussian uncertainty

    Parameters
    ----------
    mean : `float`, optional
        Mean of the Gaussian noise.

        If None, mean will be assigned a random value between 0 and 1.

        The default is None.
    scale : `float`, optional
        Scale (i.e. standard deviation) of the Gaussian noise.

        If None, scale will be assigned a random value between 0 and 1.

        The default is None.
    """
    def __init__(self, mean: float = None, scale: float = None, **kwds):
        super().__init__(**kwds)

        self._mean = mean
        self._scale = scale

    @property
    def mean(self) -> float:
        """
        Gets the mean of the Gaussian noise.

        Returns
        -------
        `float`
            Mean of the Gaussian noise.
        """
        return self._mean

    @property
    def scale(self) -> float:
        """
        Gets the scale (i.e. standard deviation) of the Gaussian noise.

        Returns
        -------
        `float`
            Scale (i.e. standard deviation) of the Gaussian noise.
        """
        return self._scale

    def get_attributes(self) -> dict:
        return super().get_attributes() | {"mean": self._mean, "scale": self._scale}

    def __eq__(self, other) -> bool:
        if not isinstance(other, GaussianUncertainty):
            raise TypeError("Can not compare 'GaussianUncertainty' instance " +
                            f"with '{type(other)}' instance")

        return super().__eq__(other) and self._mean == other.mean and self._scale == other.scale

    def __str__(self) -> str:
        return super().__str__() + f" mean: {self._mean} scale: {self._scale}"


@serializable(ABSOLUTE_GAUSSIAN_UNCERTAINTY_ID, ".epytflow_uncertainty_absolute_gaussian")
class AbsoluteGaussianUncertainty(GaussianUncertainty, JsonSerializable):
    """
    Class implementing absolute Gaussian uncertainty -- i.e. Gaussian noise is added to the data.
    """
    def apply(self, data: float) -> float:
        self._mean = self._random_generator.rand() if self._mean is None else self._mean
        self._scale = self._random_generator.rand() if self._scale is None else self._scale

        data += self._random_generator.normal(loc=self._mean, scale=self._scale)

        return self.clip(data)


@serializable(RELATIVE_GAUSSIAN_UNCERTAINTY_ID, ".epytflow_uncertainty_relative_gaussian")
class RelativeGaussianUncertainty(GaussianUncertainty, JsonSerializable):
    """
    Class implementing relative Gaussian uncertainty -- i.e. data is perturbed by Gaussian noise
    centered at zero.

    Parameters
    ----------
    scale : `float`, optional
        Scale (i.e. standard deviation) of the Gaussian noise.

        If None, scale will be assigned a random value between 0 and 1.

        The default is None.
    """
    def __init__(self, scale: float = None, **kwds):
        super().__init__(mean=0., scale=scale, **kwds)

    def apply(self, data: float) -> float:
        self._scale = self._random_generator.rand() if self._scale is None else self._scale

        data += self._random_generator.normal(loc=0, scale=self._scale)

        return self.clip(data)


class UniformUncertainty(Uncertainty):
    """
    Base class implementing uniform uncertainty.

    Parameters
    ----------
    low : `float`, optional
        Lower bound of the uniform noise.

        The default is zero.
    high : `float`, optional
        Upper bound of the uniform noise.

        The default is one.
    """
    def __init__(self, low: float = 0., high: float = 1., **kwds):
        super().__init__(**kwds)

        self.__low = low
        self.__high = high

    @property
    def low(self) -> float:
        """
        Gets the lower bound of the uniform noise.

        Returns
        -------
        `float`
            Lower bound of the uniform noise.
        """
        return self.__low

    @property
    def high(self) -> float:
        """
        Gets the upper bound of the uniform noise.

        Returns
        -------
        `float`
            Upper bound of the uniform noise.
        """
        return self.__high

    def get_attributes(self) -> dict:
        return super().get_attributes() | {"low": self.__low, "high": self.__high}

    def __eq__(self, other) -> bool:
        if not isinstance(other, UniformUncertainty):
            raise TypeError("Can not compare 'UniformUncertainty' instance " +
                            f"with '{type(other)}' instance")

        return super().__eq__(other) and self.__low == other.low and self.__high == other.high

    def __str__(self) -> str:
        return super().__str__() + f" low: {self.__low} high: {self.__high}"


@serializable(ABSOLUTE_UNIFORM_UNCERTAINTY_ID, ".epytflow_uncertainty_absolute_uniform")
class AbsoluteUniformUncertainty(UniformUncertainty, JsonSerializable):
    """
    Class implementing absolute uniform uncertainty -- i.e. uniform noise is added to the data.
    """
    def apply(self, data: float) -> float:
        data += self._random_generator.uniform(low=self.low, high=self.high)

        return self.clip(data)


@serializable(RELATIVE_UNIFORM_UNCERTAINTY_ID, ".epytflow_uncertainty_relative_uniform")
class RelativeUniformUncertainty(UniformUncertainty, JsonSerializable):
    """
    Class implementing relative uniform uncertainty -- i.e. data is multiplied by uniform noise.
    """
    def apply(self, data: float) -> float:
        data *= self._random_generator.uniform(low=self.low, high=self.high)

        return self.clip(data)


@serializable(PERCENTAGE_DEVIATON_UNCERTAINTY_ID, ".epytflow_uncertainty_percentage_deviation")
class PercentageDeviationUncertainty(UniformUncertainty, JsonSerializable):
    """
    Class implementing a uniform data deviation -- i.e. the data can deviate up to some percentage
    from its original value.

    Parameters
    ----------
    deviation_percentage : `float`
        Percentage (0-1) the data can deviate from its original value.
    """
    def __init__(self, deviation_percentage: float, **kwds):
        if not isinstance(deviation_percentage, float):
            raise TypeError("'deviation_percentage' must be an instance of 'float' " +
                            f"but not of {type(deviation_percentage)}")
        if not 0 < deviation_percentage < 1:
            raise ValueError("'deviation_percentage' must be in (0,1)")

        if "low" in kwds:
            del kwds["low"]
        if "high" in kwds:
            del kwds["high"]

        super().__init__(low=1. - deviation_percentage, high=1. + deviation_percentage, **kwds)

    def get_attributes(self) -> dict:
        return super().get_attributes() | {"deviation_percentage": self.high - 1.}

    def apply(self, data: float) -> float:
        data *= self._random_generator.uniform(low=self.low, high=self.high)

        return self.clip(data)


class DeepUniformUncertainty(Uncertainty):
    """
    Base class implementing deep uniform uncertainty.
    """
    def __init__(self, **kwds):
        super().__init__(**kwds)

        self.__create_uncertainties()

    def __create_uncertainties(self, n_samples: int = 500):
        self._uncertainties_idx = 0
        rand_low = create_deep_random_pattern(n_samples, np_rand_gen=self._random_generator)
        rand_high = create_deep_random_pattern(n_samples, np_rand_gen=self._random_generator)
        rand_low = np.minimum(rand_low, rand_high)
        rand_high = np.maximum(rand_low, rand_high)
        self._uncertainties = [self._random_generator.uniform(low, high)
                               for low, high in zip(rand_low, rand_high)]

    def set_random_generator(self, np_rand_generator) -> None:
        super().set_random_generator(np_rand_generator)

        self.__create_uncertainties()

    @abstractmethod
    def apply(self, data: float) -> float:
        self._uncertainties_idx += 1
        if self._uncertainties_idx >= len(self._uncertainties):
            self.__create_uncertainties()

        return self.clip(data)


@serializable(ABSOLUTE_DEEP_UNIFORM_UNCERTAINTY_ID, ".epytflow_uncertainty_absolute_deep_uniform")
class AbsoluteDeepUniformUncertainty(DeepUniformUncertainty, JsonSerializable):
    """
    Class implementing absolute deep uniform uncertainty -- i.e. random uniform noise
    (shape of the noise is changing over time) is added to the data.
    """
    def apply(self, data: float) -> float:
        data += self._uncertainties[self._uncertainties_idx]

        return super().apply(data)


@serializable(RELATIVE_DEEP_UNIFORM_UNCERTAINTY_ID, ".epytflow_uncertainty_relative_deep_uniform")
class RelativeDeepUniformUncertainty(DeepUniformUncertainty, JsonSerializable):
    """
    Class implementing relative deep uniform uncertainty -- i.e. data is multiplied by
    random uniform noise (shape of the noise is changing over time).
    """
    def apply(self, data: float) -> float:
        data *= self._uncertainties[self._uncertainties_idx]

        return super().apply(data)


class DeepGaussianUncertainty(Uncertainty, JsonSerializable):
    """
    Base class implementing deep Gaussian uncertainty.

    Parameters
    ----------
    mean : `float`, optional
        Fixed mean of Gaussian noise.
        If None, random means are generated.

        The default is None.
    """
    def __init__(self, mean: float = None, **kwds):
        self.__mean = mean

        super().__init__(**kwds)

        self.__create_uncertainties()

    def __create_uncertainties(self, n_samples: int = 500) -> None:
        self._uncertainties_idx = 0
        self._uncertainties = generate_deep_random_gaussian_noise(n_samples, self.__mean,
                                                                  np_rand_gen=self._random_generator)

    def set_random_generator(self, np_rand_generator) -> None:
        super().set_random_generator(np_rand_generator)

        self.__create_uncertainties()

    @abstractmethod
    def apply(self, data: float) -> float:
        self._uncertainties_idx += 1
        if self._uncertainties_idx >= len(self._uncertainties):
            self.__create_uncertainties()

        return self.clip(data)


@serializable(ABSOLUTE_DEEP_GAUSSIAN_UNCERTAINTY_ID, ".epytflow_uncertainty_absolute_deep_gaussian")
class AbsoluteDeepGaussianUncertainty(DeepGaussianUncertainty, JsonSerializable):
    """
    Class implementing absolute deep Gaussian uncertainty -- i.e. random Gaussian noise
    (mean and variance are changing over time) is added to the data.
    """
    def apply(self, data: float) -> float:
        data += self._uncertainties[self._uncertainties_idx]

        return super().apply(data)


@serializable(RELATIVE_DEEP_GAUSSIAN_UNCERTAINTY_ID, ".epytflow_uncertainty_relative_deep_gaussian")
class RelativeDeepGaussianUncertainty(DeepGaussianUncertainty, JsonSerializable):
    """
    Class implementing realtive deep Gaussian uncertainty -- i.e. data is multiplied by
    random Gaussian noise (mean and variance are changing over time).
    """
    def __init__(self, **kwds):
        super().__init__(mean=0., **kwds)

    def apply(self, data: float) -> float:
        data += self._uncertainties[self._uncertainties_idx]

        return super().apply(data)


class DeepUncertainty(Uncertainty):
    """
    Base class implementing deep uncertainty.

    Parameters
    ----------
    min_noise_value : `float`
        Lower bound on the noise.
    max_noise_value : `float`
        Upper bound on the noise.
    """
    def __init__(self, min_noise_value: float = 0., max_noise_value: float = 1., **kwds):
        super().__init__(**kwds)

        self.__min_noise_value = min_noise_value
        self.__max_noise_value = max_noise_value

        self._uncertainties_idx = None
        self._uncertainties = None
        self.__create_uncertainties()

    @property
    def min_noise_value(self) -> float:
        """
        Gets the lower bound on the noise.

        Returns
        -------
        `float`
            Lower bound on the noise.
        """
        return self.__min_noise_value

    @property
    def max_noise_value(self) -> float:
        """
        Gets the upper bound on the noise.

        Returns
        -------
        `float`
            Upper bound on the noise.
        """
        return self.__max_noise_value

    def set_random_generator(self, np_rand_generator) -> None:
        super().set_random_generator(np_rand_generator)

        self.__create_uncertainties()

    def get_attributes(self) -> dict:
        return super().get_attributes() | {"min_noise_value": self.__min_noise_value,
                                           "max_noise_value": self.__max_noise_value}

    def __eq__(self, other) -> bool:
        if not isinstance(other, DeepUncertainty):
            raise TypeError("Can not compare 'DeepUncertainty' instance " +
                            f"with '{type(other)}' instance")

        return super().__eq__(other) and self.__min_noise_value == other.min_noise_value and \
            self.__max_noise_value == other.max_noise_value

    def __str__(self) -> str:
        return super().__str__() + f" min_noise_value: {self.__min_noise_value} " +\
            f"max_noise_value: {self.__max_noise_value}"

    def __create_uncertainties(self, n_samples: int = 500) -> None:
        init_value = None
        if self._uncertainties_idx is not None:
            init_value = self._uncertainties[-1]

        self._uncertainties_idx = 0
        self._uncertainties = create_deep_random_pattern(n_samples, self.__min_noise_value,
                                                         self.__max_noise_value, init_value,
                                                         np_rand_gen=self._random_generator)

    @abstractmethod
    def apply(self, data: float) -> float:
        self._uncertainties_idx += 1
        if self._uncertainties_idx >= len(self._uncertainties):
            self.__create_uncertainties()

        return self.clip(data)


@serializable(ABSOLUTE_DEEP_UNCERTAINTY_ID, ".epytflow_uncertainty_absolute_deep")
class AbsoluteDeepUncertainty(DeepUncertainty, JsonSerializable):
    """
    Class implementing absolute deep uncertainty -- i.e. completely random noise
    is added to the data.
    """
    def apply(self, data: float) -> float:
        data += self._uncertainties[self._uncertainties_idx]

        return super().apply(data)


@serializable(RELATIVE_DEEP_UNCERTAINTY_ID, ".epytflow_uncertainty_relative_deep")
class RelativeDeepUncertainty(DeepUncertainty, JsonSerializable):
    """
    Class implementing relative deep uncertainty -- i.e. data is multiplied by
    completely random noise.
    """
    def apply(self, data: float) -> float:
        data *= self._uncertainties[self._uncertainties_idx]

        return super().apply(data)



"""
Module provides a class for implementing model uncertainty.
"""
from typing import Optional
from copy import deepcopy
import warnings
from epanet_plus import EPyT, EpanetConstants
import numpy as np

from ..serialization import serializable, JsonSerializable, MODEL_UNCERTAINTY_ID
from .uncertainties import Uncertainty


@serializable(MODEL_UNCERTAINTY_ID, ".epytflow_uncertainty_model_uncertainty")
class ModelUncertainty(JsonSerializable):
    """
    Class implementing model uncertainty -- i.e. uncertainties in pipe length, pipe roughness,
    base demand, etc.

    Parameters
    ----------
    global_pipe_length_uncertainty : :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`, optional
        Global uncertainty of pipe lengths. None, in the case of no uncertainty.

        The default is None.
    global_pipe_roughness_uncertainty : :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`, optional
        Global uncertainty of pipe roughness coefficients. None, in the case of no uncertainty.

        The default is None.
    global_pipe_diameter_uncertainty : :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`, optional
        Global uncertainty of pipe diameters. None, in the case of no uncertainty.

        The default is None.
    global_base_demand_uncertainty : :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`, optional
        Global uncertainty of base demands. None, in the case of no uncertainty.

        The default is None.
    global_demand_pattern_uncertainty : :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`, optional
        Global uncertainty of demand patterns. None, in the case of no uncertainty.

        The default is None.
    global_elevation_uncertainty : :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`, optional
        Global uncertainty of elevations. None, in the case of no uncertainty.

        The default is None.
    global_constants_uncertainty : :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`, optional
        Global uncertainty of MSX constants. None, in the case of no uncertainty.

        The default is None.
    global_parameters_uncertainty : :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`, optional
        Global uncertainty of MSX parameters. None, in the case of no uncertaint.

        The default is None.
    local_pipe_length_uncertainty : dict[str, :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`], optional
        Local uncertainty of pipe lengths -- i.e. a dictionary of pipe IDs and uncertainties.

        None, in the case of no uncertainty.

        The default is None.
    local_pipe_roughness_uncertainty : dict[str, :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`], optional
        Local uncertainty of pipe roughness coefficients -- i.e. a dictionary of pipe IDs and uncertainties.

        None, in the case of no uncertainty.

        The default is None.
    local_pipe_diameter_uncertainty : dict[str, :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`], optional
        Local uncertainty of pipe diameters -- i.e. a dictionary of pipe IDs and uncertainties.

        None, in the case of no uncertainty.

        The default is None.
    local_base_demand_uncertainty : dict[str, :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`], optional
        Local uncertainty of base demands -- i.e. a dictionary of node IDs and uncertainties.

        None, in the case of no uncertainty.

        The default is None.
    local_demand_pattern_uncertainty : dict[str, :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`], optional
        Local uncertainty of demand patterns --
        i.e. a dictionary of demand pattern IDs and uncertainties.

        None, in the case of no uncertainty.

        The default is None.
    local_elevation_uncertainty : dict[str, :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`], optional
        Local uncertainty of elevations -- i.e. a dictionary of node IDs and uncertainties.

        None, in the case of no uncertainty.

        The default is None.
    local_constants_uncertainty : dict[str, :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`], optional
        Local uncertainty of MSX constants -- i.e. a dictionary of constant IDs and uncertainties.

        None, in the case of no uncertainty.

        The default is None.
    local_parameters_uncertainty : dict[tuple[str, int, str] :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`], optional
        Local uncertainty of MSX parameters -- i.e. a dictionary of
        (parameter ID, item type, item ID) and uncertainties.

        None, in the case of no uncertainty.

        The default is None.
    local_patterns_uncertainty : dict[str, :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`], optional
        Local uncertainty of EPANET patterns -- i.e. a dictionary of pattern IDs and uncertainties.

        None, in the case of no uncertainty.

        The default is None.
    local_msx_patterns_uncertainty : dict[str, :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`], optional
        Local uncertainty of EPANET-MSX patterns -- i.e. a dictionary of MSX pattern IDs
        and uncertainties.

        None, in the case of no uncertainty.

        The default is None.
    seed : `int`, optional
        Seed for the random number generator.

        Thed default is None.
    cache_original : `bool`, optional
        If True, all original properties are cached before the uncertainties are applied.
        This is necessary if you have multiple simulation runs and you want to reset/re-apply
        the uncertainties before each run.

        You can set it to False if you do not and want to re-apply the uncertainties and
        save some working memory.

        The default is True.
    """
    def __init__(self, global_pipe_length_uncertainty: Optional[Uncertainty] = None,
                 global_pipe_roughness_uncertainty: Optional[Uncertainty] = None,
                 global_pipe_diameter_uncertainty: Optional[Uncertainty] = None,
                 global_base_demand_uncertainty: Optional[Uncertainty] = None,
                 global_demand_pattern_uncertainty: Optional[Uncertainty] = None,
                 global_elevation_uncertainty: Optional[Uncertainty] = None,
                 global_constants_uncertainty: Optional[Uncertainty] = None,
                 global_parameters_uncertainty: Optional[Uncertainty] = None,
                 local_pipe_length_uncertainty: Optional[dict[str, Uncertainty]] = None,
                 local_pipe_roughness_uncertainty: Optional[dict[str, Uncertainty]] = None,
                 local_pipe_diameter_uncertainty: Optional[dict[str, Uncertainty]] = None,
                 local_base_demand_uncertainty: Optional[dict[str, Uncertainty]] = None,
                 local_demand_pattern_uncertainty: Optional[dict[str, Uncertainty]] = None,
                 local_elevation_uncertainty: Optional[dict[str, Uncertainty]] = None,
                 local_constants_uncertainty: Optional[dict[str, Uncertainty]] = None,
                 local_parameters_uncertainty: Optional[dict[str, int, Uncertainty]] = None,
                 local_patterns_uncertainty: Optional[dict[str, Uncertainty]] = None,
                 local_msx_patterns_uncertainty: Optional[dict[str, Uncertainty]] = None,
                 seed: Optional[int] = None, cache_original: Optional[bool] = True,
                 **kwds):
        if global_pipe_length_uncertainty is not None:
            if not isinstance(global_pipe_length_uncertainty, Uncertainty):
                raise TypeError("'global_pipe_length_uncertainty' must be an instance of " +
                                "'epyt_flow.uncertainty.Uncertainty' but not of " +
                                f"'{type(global_pipe_length_uncertainty)}'")
        if global_pipe_roughness_uncertainty is not None:
            if not isinstance(global_pipe_roughness_uncertainty, Uncertainty):
                raise TypeError("'global_pipe_roughness_uncertainty' must be an instance of " +
                                "'epyt_flow.uncertainty.Uncertainty' but not of " +
                                f"'{type(global_pipe_roughness_uncertainty)}'")
        if global_pipe_diameter_uncertainty is not None:
            if not isinstance(global_pipe_diameter_uncertainty, Uncertainty):
                raise TypeError("'global_pipe_diameter_uncertainty' must be an instance of " +
                                "'epyt_flow.uncertainty.Uncertainty' but not of " +
                                f"'{type(global_pipe_diameter_uncertainty)}'")
        if global_base_demand_uncertainty is not None:
            if not isinstance(global_base_demand_uncertainty, Uncertainty):
                raise TypeError("'global_base_demand_uncertainty' must be an instance of " +
                                "'epyt_flow.uncertainty.Uncertainty' but not of " +
                                f"'{type(global_base_demand_uncertainty)}'")
        if global_demand_pattern_uncertainty is not None:
            if not isinstance(global_demand_pattern_uncertainty, Uncertainty):
                raise TypeError("'global_demand_pattern_uncertainty' must be an instance of " +
                                "'epyt_flow.uncertainty.Uncertainty' but not of " +
                                f"'{type(global_demand_pattern_uncertainty)}'")
        if global_elevation_uncertainty is not None:
            if not isinstance(global_elevation_uncertainty, Uncertainty):
                raise TypeError("'global_elevation_uncertainty' must be an instance of " +
                                "'epyt_flow.uncertainty.Uncertainty' but not of " +
                                f"'{type(global_elevation_uncertainty)}'")
        if global_constants_uncertainty is not None:
            if not isinstance(global_constants_uncertainty, Uncertainty):
                raise TypeError("'global_constants_uncertainty' must be an instance of " +
                                "'epyt_flow.uncertainty.Uncertainty' but not of " +
                                f"'{type(global_constants_uncertainty)}'")
        if global_parameters_uncertainty is not None:
            if not isinstance(global_parameters_uncertainty, Uncertainty):
                raise TypeError("'global_parameters_uncertainty' must be an instance of " +
                                "'epyt_flow.uncertainty.Uncertainty' but not of " +
                                f"'{type(global_parameters_uncertainty)}'")

        if local_pipe_length_uncertainty is not None:
            if not isinstance(local_pipe_length_uncertainty, dict):
                raise TypeError("'local_pipe_length_uncertainty' must be an instance of " +
                                "'dict[str, epyt_flow.uncertainty.Uncertainty]' but not of " +
                                f"'{type(local_pipe_length_uncertainty)}'")
            if any(not isinstance(key, str) or not isinstance(val, Uncertainty)
                   for key, val in local_pipe_length_uncertainty.items()):
                raise TypeError("'local_pipe_length_uncertainty': " +
                                "All keys must be instances of 'str' and all values must be " +
                                "instances of 'epyt_flow.uncertainty.Uncertainty'")
        if local_pipe_roughness_uncertainty is not None:
            if not isinstance(local_pipe_roughness_uncertainty, dict):
                raise TypeError("'local_pipe_roughness_uncertainty' must be an instance of " +
                                "'dict[str, epyt_flow.uncertainty.Uncertainty]' but not of " +
                                f"'{type(local_pipe_roughness_uncertainty)}'")
            if any(not isinstance(key, str) or not isinstance(val, Uncertainty)
                   for key, val in local_pipe_roughness_uncertainty.items()):
                raise TypeError("'local_pipe_roughness_uncertainty': " +
                                "All keys must be instances of 'str' and all values must be " +
                                "instances of 'epyt_flow.uncertainty.Uncertainty'")
        if local_pipe_diameter_uncertainty is not None:
            if not isinstance(local_pipe_diameter_uncertainty, dict):
                raise TypeError("'local_pipe_diameter_uncertainty' must be an instance of " +
                                "'dict[str, epyt_flow.uncertainty.Uncertainty]' but not of " +
                                f"'{type(local_pipe_diameter_uncertainty)}'")
            if any(not isinstance(key, str) or not isinstance(val, Uncertainty)
                   for key, val in local_pipe_diameter_uncertainty.items()):
                raise TypeError("'local_pipe_diameter_uncertainty': " +
                                "All keys must be instances of 'str' and all values must be " +
                                "instances of 'epyt_flow.uncertainty.Uncertainty'")
        if local_base_demand_uncertainty is not None:
            if not isinstance(local_base_demand_uncertainty, dict):
                raise TypeError("'local_base_demand_uncertainty' must be an instance of " +
                                "'dict[str, epyt_flow.uncertainty.Uncertainty]' but not of " +
                                f"'{type(local_base_demand_uncertainty)}'")
            if any(not isinstance(key, str) or not isinstance(val, Uncertainty)
                   for key, val in local_base_demand_uncertainty.items()):
                raise TypeError("'local_base_demand_uncertainty': " +
                                "All keys must be instances of 'str' and all values must be " +
                                "instances of 'epyt_flow.uncertainty.Uncertainty'")
        if local_demand_pattern_uncertainty is not None:
            if not isinstance(local_demand_pattern_uncertainty, dict):
                raise TypeError("'local_demand_pattern_uncertainty' must be an instance of " +
                                "'dict[str, epyt_flow.uncertainty.Uncertainty]' but not of " +
                                f"'{type(local_demand_pattern_uncertainty)}'")
            if any(not isinstance(key, str) or not isinstance(val, Uncertainty)
                   for key, val in local_demand_pattern_uncertainty.items()):
                raise TypeError("'local_demand_pattern_uncertainty': " +
                                "All keys must be instances of 'str' and all values must be " +
                                "instances of 'epyt_flow.uncertainty.Uncertainty'")
        if local_elevation_uncertainty is not None:
            if not isinstance(local_elevation_uncertainty, dict):
                raise TypeError("'local_elevation_uncertainty' must be an instance of " +
                                "'dict[str, epyt_flow.uncertainty.Uncertainty]' but not of " +
                                f"'{type(local_elevation_uncertainty)}'")
            if any(not isinstance(key, str) or not isinstance(val, Uncertainty)
                   for key, val in local_elevation_uncertainty.items()):
                raise TypeError("'local_elevation_uncertainty': " +
                                "All keys must be instances of 'str' and all values must be " +
                                "instances of 'epyt_flow.uncertainty.Uncertainty'")
        if local_constants_uncertainty is not None:
            if not isinstance(local_constants_uncertainty, dict):
                raise TypeError("'local_constants_uncertainty' must be an instance of " +
                                "'dict[str, epyt_flow.uncertainty.Uncertainty]' but not of " +
                                f"'{type(local_constants_uncertainty)}'")
            if any(not isinstance(key, str) or not isinstance(val, Uncertainty)
                   for key, val in local_constants_uncertainty.items()):
                raise TypeError("'local_constants_uncertainty': " +
                                "All keys must be instances of 'str' and all values must be " +
                                "instances of 'epyt_flow.uncertainty.Uncertainty'")
        if local_parameters_uncertainty is not None:
            if not isinstance(local_parameters_uncertainty, dict):
                raise TypeError("'local_parameters_uncertainty' must be an instance of " +
                                "'dict[str, epyt_flow.uncertainty.Uncertainty]' but not of " +
                                f"'{type(local_parameters_uncertainty)}'")
            if any(not isinstance(key, tuple) or not isinstance(key[0], str) or
                   not isinstance(key[1], int) or not isinstance(key[2], str) or
                   not isinstance(local_parameters_uncertainty[key], Uncertainty)
                   for key in local_parameters_uncertainty.keys()):
                raise TypeError("'local_parameters_uncertainty': " +
                                "All keys must be instances of 'tuple[str, int, str]' and all " +
                                "values must be instances of 'epyt_flow.uncertainty.Uncertainty'")
        if local_patterns_uncertainty is not None:
            if not isinstance(local_patterns_uncertainty, dict):
                raise TypeError("'local_patterns_uncertainty' must be an instance of " +
                                "'dict[str, epyt_flow.uncertainty.Uncertainty]' but not of " +
                                f"'{type(local_patterns_uncertainty)}'")
            if any(not isinstance(key, str) or not isinstance(val, Uncertainty)
                   for key, val in local_patterns_uncertainty.items()):
                raise TypeError("'local_patterns_uncertainty': " +
                                "All keys must be instances of 'str' and all values must be " +
                                "instances of 'epyt_flow.uncertainty.Uncertainty'")
        if local_msx_patterns_uncertainty is not None:
            if not isinstance(local_msx_patterns_uncertainty, dict):
                raise TypeError("'local_msx_patterns_uncertainty' must be an instance of " +
                                "'dict[str, epyt_flow.uncertainty.Uncertainty]' but not of " +
                                f"'{type(local_msx_patterns_uncertainty)}'")
            if any(not isinstance(key, str) or not isinstance(val, Uncertainty)
                   for key, val in local_msx_patterns_uncertainty.items()):
                raise TypeError("'local_msx_patterns_uncertainty': " +
                                "All keys must be instances of 'str' and all values must be " +
                                "instances of 'epyt_flow.uncertainty.Uncertainty'")
        if not isinstance(cache_original, bool):
            raise TypeError("'cache_original' must be an instance of 'bool' " +
                            f"but not of '{type(cache_original)}'")

        self._global_pipe_length = global_pipe_length_uncertainty
        self._global_pipe_roughness = global_pipe_roughness_uncertainty
        self._global_pipe_diameter = global_pipe_diameter_uncertainty
        self._global_base_demand = global_base_demand_uncertainty
        self._global_demand_pattern = global_demand_pattern_uncertainty
        self._global_elevation = global_elevation_uncertainty
        self._global_constants = global_constants_uncertainty
        self._global_parameters = global_parameters_uncertainty
        self._local_pipe_length = local_pipe_length_uncertainty
        self._local_pipe_roughness = local_pipe_roughness_uncertainty
        self._local_pipe_diameter = local_pipe_diameter_uncertainty
        self._local_base_demand = local_base_demand_uncertainty
        self._local_demand_pattern = local_demand_pattern_uncertainty
        self._local_elevation = local_elevation_uncertainty
        self._local_constants = local_constants_uncertainty
        self._local_parameters = local_parameters_uncertainty
        self._local_patterns = local_patterns_uncertainty
        self._local_msx_patterns = local_msx_patterns_uncertainty
        self.__seed = seed
        self.__cache_original = cache_original

        self._cache_links_length = None
        self._cache_links_diameter = None
        self._cache_links_roughness_coeff = None
        self._cache_nodes_base_demand = None
        self._cache_nodes_demand_pattern = None
        self._cache_nodes_elevation = None
        self._cache_patterns = None
        self._cache_msx_constants = None
        self._cache_msx_links_parameters = None
        self._cache_msx_tanks_parameters = None
        self._cache_msx_patterns = None

        super().__init__(**kwds)

    @property
    def seed(self) -> int:
        """
        Returns the seed used for the random number generator.

        Returns
        -------
        `int`
            Seed for the random number generator.
        """
        return self.__seed

    @property
    def global_pipe_length(self) -> Uncertainty:
        """
        Returns the global pipe length uncertainty.

        Returns
        -------
        :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`
            Global pipe length uncertainty.
        """
        return deepcopy(self._global_pipe_length)

    @property
    def global_pipe_roughness(self) -> Uncertainty:
        """
        Returns the global pipe roughness uncertainty.

        Returns
        -------
        :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`
            Global pipe roughness uncertainty.
        """
        return deepcopy(self._global_pipe_roughness)

    @property
    def global_pipe_diameter(self) -> Uncertainty:
        """
        Returns the global pipe diameter uncertainty.

        Returns
        -------
        :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`
            Global pipe diameter uncertainty.
        """
        return deepcopy(self._global_pipe_diameter)

    @property
    def global_base_demand(self) -> Uncertainty:
        """
        Returns the global base demand uncertainty.

        Returns
        -------
        :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`
            Global base demand uncertainty.
        """
        return deepcopy(self._global_base_demand)

    @property
    def global_demand_pattern(self) -> Uncertainty:
        """
        Returns the global demand pattern uncertainty.

        Returns
        -------
        :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`
            Global demand pattern uncertainty.
        """
        return deepcopy(self._global_demand_pattern)

    @property
    def global_elevation(self) -> Uncertainty:
        """
        Returns the global node elevation uncertainty.

        Returns
        -------
        :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`
            Global node elevation uncertainty.
        """
        return deepcopy(self._global_elevation)

    @property
    def global_constants(self) -> Uncertainty:
        """
        Returns the global MSX constant uncertainty.

        Returns
        -------
        :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`
            Global MSX constant uncertainty.
        """
        return deepcopy(self._global_constants)

    @property
    def global_parameters(self) -> Uncertainty:
        """
        Returns the global MSX parameter uncertainty.

        Returns
        -------
        :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`
            Global MSX parameter uncertainty.
        """
        return deepcopy(self._global_parameters)

    @property
    def local_pipe_length(self) -> dict[str, Uncertainty]:
        """
        Returns the local pipe length uncertainty.

        Returns
        -------
        dict[str, :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`]
            Local pipe length uncertainty.
        """
        return deepcopy(self._local_pipe_length)

    @property
    def local_pipe_roughness(self) -> dict[str, Uncertainty]:
        """
        Returns the local pipe roughness uncertainty.

        Returns
        -------
        dict[str, :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`]
            Local pipe roughness uncertainty.
        """
        return deepcopy(self._local_pipe_roughness)

    @property
    def local_pipe_diameter(self) -> dict[str, Uncertainty]:
        """
        Returns the local pipe diameter uncertainty.

        Returns
        -------
        dict[str, :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`]
            Local pipe diameter uncertainty.
        """
        return deepcopy(self._local_pipe_diameter)

    @property
    def local_base_demand(self) -> dict[str, Uncertainty]:
        """
        Returns the local base demand uncertainty.

        Returns
        -------
        dict[str, :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`]
            Local base demand uncertainty.
        """
        return deepcopy(self._local_base_demand)

    @property
    def local_demand_pattern(self) -> dict[str, Uncertainty]:
        """
        Returns the local demand pattern uncertainty.

        Returns
        -------
        dict[str, :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`]
            Local demand pattern uncertainty.
        """
        return deepcopy(self._local_demand_pattern)

    @property
    def local_elevation(self) -> dict[str, Uncertainty]:
        """
        Returns the local node elevation uncertainty.

        Returns
        -------
        dict[str, :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`]
            Local node elevation uncertainty.
        """
        return deepcopy(self._local_elevation)

    @property
    def local_constants(self) -> dict[str, Uncertainty]:
        """
        Returns the local MSX constant uncertainty.

        Returns
        -------
        dict[str, :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`]
            Local MSX constant uncertainty.
        """
        return deepcopy(self._local_constants)

    @property
    def local_parameters(self) -> dict[tuple[str, int, str], Uncertainty]:
        """
        Returns the local MSX parameter uncertainty.

        Returns
        -------
        dict[tuple[str, int, str], :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`]
            Local MSX parameter uncertainty.
        """
        return deepcopy(self._local_parameters)

    @property
    def local_patterns(self) -> dict[str, Uncertainty]:
        """
        Returns the local EPANET patterns uncertainty.

        Returns
        -------
        dict[str, :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`]
            Local EPANET patterns uncertainty.
        """
        return deepcopy(self._local_patterns)

    @property
    def local_msx_patterns(self) -> dict[str, Uncertainty]:
        """
        Returns the local EPANET-MSX patterns uncertainty.

        Returns
        -------
        dict[str, :class:`~epyt_flow.uncertainty.uncertainties.Uncertainty`]
            Local EPANET-MSX patterns uncertainty.
        """
        return deepcopy(self._local_msx_patterns)

    def get_attributes(self) -> dict:
        attribs = {"global_pipe_length_uncertainty": self._global_pipe_length,
                   "global_pipe_roughness_uncertainty": self._global_pipe_roughness,
                   "global_pipe_diameter_uncertainty": self._global_pipe_diameter,
                   "global_base_demand_uncertainty": self._global_base_demand,
                   "global_demand_pattern_uncertainty": self._global_demand_pattern,
                   "global_elevation_uncertainty": self._global_elevation,
                   "global_constants_uncertainty": self._global_constants,
                   "global_parameters_uncertainty": self._global_parameters,
                   "local_pipe_length_uncertainty": self._local_pipe_length,
                   "local_pipe_roughness_uncertainty": self._local_pipe_roughness,
                   "local_pipe_diameter_uncertainty": self._local_pipe_diameter,
                   "local_base_demand_uncertainty": self._local_base_demand,
                   "local_demand_pattern_uncertainty": self._local_demand_pattern,
                   "local_elevation_uncertainty": self._local_elevation,
                   "local_constants_uncertainty": self._local_constants,
                   "local_parameters_uncertainty": self._local_parameters,
                   "local_patterns_uncertainty": self._local_patterns,
                   "local_msx_patterns_uncertainty": self._local_msx_patterns,
                   "seed": self.__seed}

        return super().get_attributes() | attribs

    def __eq__(self, other) -> bool:
        if not isinstance(other, ModelUncertainty):
            raise TypeError("Can not compare 'ModelUncertainty' instance " +
                            f"with '{type(other)}' instance")

        return self._global_pipe_length == other.global_pipe_length \
            and self._global_pipe_roughness == other.global_pipe_roughness \
            and self._global_pipe_diameter == other.global_pipe_diameter \
            and self._global_base_demand == other.global_base_demand \
            and self._global_demand_pattern == other.global_demand_pattern \
            and self._global_elevation == other.global_elevation \
            and self._global_parameters == other.global_parameters \
            and self._global_constants == other.global_constants \
            and self._local_pipe_length == other.local_pipe_length \
            and self._local_pipe_roughness == other.local_pipe_roughness \
            and self._local_pipe_diameter == other.local_pipe_diameter \
            and self._local_base_demand == other.local_base_demand \
            and self._local_demand_pattern == other.local_demand_pattern \
            and self._local_elevation == other.local_elevation \
            and self._local_parameters == other.local_parameters \
            and self._local_constants == other.local_constants \
            and self._local_patterns == other.local_patterns \
            and self._local_msx_patterns == other.local_msx_patterns \
            and self.__seed == other.seed

    def __str__(self) -> str:
        return f"global_pipe_length: {self._global_pipe_length} " +\
            f"global_pipe_roughness: {self._global_pipe_roughness} " + \
            f"global_pipe_diameter: {self._global_pipe_diameter} " + \
            f"global_demand_base: {self._global_base_demand} " + \
            f"global_demand_pattern: {self._global_demand_pattern} " + \
            f"global_elevation: {self._global_elevation} " + \
            f"global_constants: {self._global_constants} " + \
            f"global_parameters: {self._global_parameters}" + \
            f"local_pipe_length: {self._local_pipe_length} " +\
            f"local_pipe_roughness: {self._local_pipe_roughness} " + \
            f"local_pipe_diameter: {self._local_pipe_diameter} " + \
            f"local_demand_base: {self._local_base_demand} " + \
            f"local_demand_pattern: {self._local_demand_pattern} " + \
            f"local_elevation: {self._local_elevation} " + \
            f"local_constants: {self._local_constants} " + \
            f"local_parameters: {self._local_parameters} " + \
            f"local_patterns: {self._local_patterns} " + \
            f"local_msx_patterns: {self._local_msx_patterns} + seed: {self.__seed}"

    def undo(self, epanet_api: EPyT) -> None:
        """
        Undo all applied uncertainties -- i.e, resets the properties to their original value.

        Note that this function can only be used if `cache_original` (of the constructor)
        was set to True (default).

        Parameters
        ----------
        epanet_api : `epanet_plus.EPyT <https://epanet-plus.readthedocs.io/en/stable/api.html#epanet_plus.epanet_toolkit.EPyT>`_
            Interface to EPANET and EPANET-MSX
        """
        if self.__cache_original is False:
            raise ValueError("Caching was disabled by the user")

        if self._cache_links_length is not None:
            for link_idx, link_len in zip(epanet_api.get_all_links_idx(),
                                          self._cache_links_length):
                epanet_api.setlinkvalue(link_idx, EpanetConstants.EN_LENGTH, link_len)

        if self._cache_links_diameter is not None:
            for link_idx, link_diam in zip(epanet_api.get_all_links_idx(),
                                           self._cache_links_diameter):
                epanet_api.setlinkvalue(link_idx, EpanetConstants.EN_DIAMETER, link_diam)

        if self._cache_links_roughness_coeff is not None:
            for link_idx, link_roughness in zip(epanet_api.get_all_links_idx(),
                                                self._cache_links_roughness_coeff):
                epanet_api.setlinkvalue(link_idx, EpanetConstants.EN_ROUGHNESS, link_roughness)

        if self._cache_nodes_base_demand is not None:
            for node_idx in self._cache_nodes_base_demand.keys():
                for demand_category, base_demand in self._cache_nodes_base_demand[node_idx].items():
                    epanet_api.setbasedemand(node_idx, demand_category + 1, base_demand)

        if self._cache_nodes_demand_pattern is not None:
            for pattern_idx, demand_pattern in self._cache_nodes_demand_pattern.items():
                epanet_api.set_pattern(pattern_idx, demand_pattern.tolist())

        if self._cache_nodes_elevation is not None:
            for node_idx, node_elev in zip(epanet_api.get_all_nodes_idx(),
                                           self._cache_nodes_elevation):
                epanet_api.setnodevalue(node_idx, EpanetConstants.EN_ELEVATION, node_elev)

        if self._cache_patterns is not None:
            for pattern_idx, pattern in self._cache_patterns.items():
                epanet_api.set_pattern(pattern_idx, pattern.tolist())

        if self._cache_msx_constants is not None:
            for constant_idx, constant_value in enumerate(self._cache_msx_constants):
                epanet_api.MSXsetconstant(constant_idx + 1, constant_value)

        if self._cache_msx_links_parameters is not None:
            for pipe_idx, parameters_pipes_val in self._cache_msx_links_parameters.items():
                for param_idx, param_value in enumerate(parameters_pipes_val):
                    epanet_api.MSXsetparameter(EpanetConstants.MSX_LINK, pipe_idx,
                                               param_idx + 1, param_value)

        if self._cache_msx_tanks_parameters is not None:
            for tank_idx, parameters_tanks_val in self._cache_msx_tanks_parameters.items():
                for param_idx, param_value in enumerate(parameters_tanks_val):
                    epanet_api.MSXsetparameter(EpanetConstants.MSX_NODE, tank_idx,
                                               param_idx + 1, param_value)

        if self._cache_msx_patterns is not None:
            for pattern_idx, pattern in self._cache_msx_patterns:
                epanet_api.MSXsetpattern(pattern_idx, pattern.tolist(), len(pattern))

    def apply(self, epanet_api: EPyT) -> None:
        """
        Applies the specified model uncertainties to the scenario.

        Parameters
        ----------
        epanet_api : `epanet_plus.EPyT <https://epanet-plus.readthedocs.io/en/stable/api.html#epanet_plus.epanet_toolkit.EPyT>`_
            Interface to EPANET and EPANET-MSX.
        """
        np_rand_gen = np.random.default_rng(seed=self.__seed)

        all_links_idx = epanet_api.get_all_links_idx()
        all_nodes_idx = epanet_api.get_all_nodes_idx()

        if self._global_pipe_length is not None:
            self._global_pipe_length.set_random_generator(np_rand_gen)

            link_length = np.array([epanet_api.getlinkvalue(link_idx, EpanetConstants.EN_LENGTH)
                                    for link_idx in all_links_idx])
            self._cache_links_length = np.copy(link_length)

            link_length = self._global_pipe_length.apply_batch(link_length)
            for link_idx, link_value in zip(all_links_idx, link_length):
                epanet_api.setlinkvalue(link_idx, EpanetConstants.EN_LENGTH, link_value)

        if self._local_pipe_length is not None:
            self._local_pipe_length.set_random_generator(np_rand_gen)
            self._cache_links_length = np.array([epanet_api.getlinkvalue(link_idx, EpanetConstants.EN_LENGTH)
                                                 for link_idx in all_links_idx])

            for pipe_id, uncertainty in self._local_pipe_length.items():
                link_idx = epanet_api.get_link_idx(pipe_id)
                link_length = epanet_api.getlinkvalue(link_idx, EpanetConstants.EN_LENGTH)

                link_length = uncertainty.apply(link_length)
                epanet_api.setlinkvalue(link_idx, EpanetConstants.EN_LENGTH, link_length)

        if self._global_pipe_diameter is not None:
            self._global_pipe_diameter.set_random_generator(np_rand_gen)

            link_diameters = np.array([epanet_api.getlinkvalue(link_idx, EpanetConstants.EN_DIAMETER)
                                       for link_idx in all_links_idx])
            self._cache_links_diameter = np.copy(link_diameters)

            link_diameters = self._global_pipe_diameter.apply_batch(link_diameters)
            for link_idx, link_value in zip(all_links_idx, link_diameters):
                epanet_api.setlinkvalue(link_idx, EpanetConstants.EN_DIAMETER, link_value)

        if self._local_pipe_diameter is not None:
            self._local_pipe_diameter.set_random_generator(np_rand_gen)
            self._cache_links_diameter = np.array([epanet_api.getlinkvalue(link_idx,
                                                                           EpanetConstants.EN_DIAMETER)
                                                   for link_idx in all_links_idx])

            for pipe_id, uncertainty in self._local_pipe_diameter.items():
                link_idx = epanet_api.get_link_idx(pipe_id)
                link_diameter = epanet_api.getlinkvalue(link_idx, EpanetConstants.EN_DIAMETER)

                link_diameter = uncertainty.apply(link_diameter)
                epanet_api.setlinkvalue(link_idx, EpanetConstants.EN_DIAMETER, link_diameter)

        if self._global_pipe_roughness is not None:
            self._global_pipe_roughness.set_random_generator(np_rand_gen)

            coeffs = np.array([epanet_api.getlinkvalue(link_idx, EpanetConstants.EN_ROUGHNESS)
                               for link_idx in all_links_idx])
            self._cache_links_roughness_coeff = np.copy(coeffs)

            coeffs = self._global_pipe_roughness.apply_batch(coeffs)
            for link_idx, link_value in zip(all_links_idx, coeffs):
                epanet_api.setlinkvalue(link_idx, EpanetConstants.EN_ROUGHNESS, link_value)

        if self._local_pipe_roughness is not None:
            self._local_pipe_roughness.set_random_generator(np_rand_gen)
            self._cache_links_roughness_coeff = \
                np.array([epanet_api.getlinkvalue(link_idx, EpanetConstants.EN_ROUGHNESS)
                          for link_idx in all_links_idx])

            for pipe_id, uncertainty in self._local_pipe_roughness.items():
                link_idx = epanet_api.get_link_idx(pipe_id)
                link_roughness_coeff = epanet_api.getlinkvalue(link_idx,
                                                               EpanetConstants.EN_ROUGHNESS)

                link_roughness_coeff = uncertainty.apply(link_roughness_coeff)
                epanet_api.setlinkvalue(link_idx, EpanetConstants.EN_ROUGHNESS,
                                        link_roughness_coeff)

        if self._global_base_demand is not None:
            self._global_base_demand.set_random_generator(np_rand_gen)

            self._cache_nodes_base_demand = {}
            for node_idx in all_nodes_idx:
                self._cache_nodes_base_demand[node_idx] = {}
                n_demand_categories = epanet_api.getnumdemands(node_idx)
                for demand_idx in range(n_demand_categories):
                    base_demand = epanet_api.getbasedemand(node_idx, demand_idx + 1)
                    self._cache_nodes_base_demand[node_idx][demand_idx] = base_demand

                    base_demand = self._global_base_demand.apply(base_demand)
                    epanet_api.setbasedemand(node_idx, demand_idx + 1, base_demand)

        if self._local_base_demand is not None:
            self._local_base_demand.set_random_generator(np_rand_gen)

            self._cache_nodes_base_demand = {}
            for node_id, uncertainty in self._local_base_demand.items():
                node_idx = epanet_api.get_node_idx(node_id)
                self._cache_nodes_base_demand[node_idx] = {}
                n_demand_categories = epanet_api.getnumdemands(node_idx)
                for demand_idx in range(n_demand_categories):
                    base_demand = epanet_api.getbasedemand(node_idx, demand_idx + 1)
                    self._cache_nodes_base_demand[node_idx][demand_idx] = base_demand

                    base_demand = uncertainty.apply(base_demand)
                    epanet_api.setbasedemand(node_idx, demand_idx + 1, base_demand)

        if self._global_demand_pattern is not None:
            self._global_demand_pattern.set_random_generator(np_rand_gen)

            self._cache_nodes_demand_pattern = {}
            demand_patterns_idx = np.array([epanet_api.getdemandpattern(node_idx, demand_idx + 1)
                                            for node_idx in epanet_api.get_all_nodes_idx()
                                            for demand_idx in range(epanet_api.getnumdemands(node_idx))])

            for pattern_idx in list(set(demand_patterns_idx)):
                demand_pattern = np.array(epanet_api.get_pattern(pattern_idx))
                self._cache_nodes_demand_pattern[pattern_idx] = np.copy(demand_pattern)

                demand_pattern = self._global_demand_pattern.apply_batch(demand_pattern)
                epanet_api.set_pattern(pattern_idx, demand_pattern.tolist())

        if self._local_demand_pattern is not None:
            self._local_demand_pattern.set_random_generator(np_rand_gen)

            self._cache_nodes_demand_pattern = {}

            for pattern_id, uncertainty in self._local_demand_pattern.items():
                pattern_idx = epanet_api.getpatternindex(pattern_id)
                demand_pattern = np.array(epanet_api.get_pattern(pattern_idx))
                self._cache_nodes_demand_pattern[pattern_id] = np.copy(demand_pattern)

                demand_pattern = uncertainty.apply_batch(demand_pattern)
                epanet_api.set_pattern(pattern_idx, demand_pattern.tolist())

        if self._global_elevation is not None:
            self._global_elevation.set_random_generator(np_rand_gen)

            elevations = np.array([epanet_api.get_node_elevation(node_idx)
                                   for node_idx in epanet_api.get_all_nodes_idx()])
            self._cache_nodes_elevation = np.copy(elevations)

            elevations = self._global_elevation.apply_batch(elevations)
            for node_idx, node_elev in enumerate(elevations):
                epanet_api.setnodevalue(node_idx + 1, EpanetConstants.EN_ELEVATION, node_elev)

        if self._local_elevation is not None:
            self._local_elevation.set_random_generator(np_rand_gen)
            self._cache_nodes_elevation = np.array([epanet_api.get_node_elevation(node_idx)
                                                    for node_idx in epanet_api.get_all_nodes_idx()])

            for node_id, uncertainty in self._local_elevation.items():
                node_idx = epanet_api.get_node_idx(node_id)
                elevation = epanet_api.get_node_elevation(node_idx)

                elevation = uncertainty.apply(elevation)
                epanet_api.setnodevalue(node_idx, EpanetConstants.EN_ELEVATION, elevation)

        if self._local_patterns is not None:
            self._local_patterns.set_random_generator(np_rand_gen)
            self._cache_patterns = {}

            for pattern_id, uncertainty in self._local_patterns.items():
                pattern_idx = epanet_api.getpatternindex(pattern_id)
                pattern = np.array(epanet_api.get_pattern(pattern_idx))
                self._cache_patterns[pattern_idx] = np.copy(pattern)

                pattern = uncertainty.apply_batch(pattern)
                epanet_api.set_pattern(pattern_idx, pattern.tolist())

        if epanet_api.msx_file is not None:
            if self._global_constants is not None:
                self._global_constants.set_random_generator(np_rand_gen)

                constants = np.array([epanet_api.MSXgetconstant(const_idx + 1)
                                      for const_idx in range(epanet_api.MSXgetcount(EpanetConstants.MSX_CONSTANT))])
                self._cache_msx_patterns = np.copy(constants)

                constants = self._global_constants.apply_batch(constants)
                for const_idx, const_value in enumerate(constants):
                    epanet_api.MSXsetconstant(const_idx + 1, const_value)

            if self._local_constants:
                self._local_constants.set_random_generator(np_rand_gen)

                self._cache_msx_patterns = np.array([epanet_api.MSXgetconstant(const_idx + 1)
                                                     for const_idx in range(epanet_api.MSXgetcount(EpanetConstants.MSX_CONSTANT))])

                for constant_id, uncertainty in self._local_constants.items():
                    idx = epanet_api.MSXgetindex(EpanetConstants.MSX_CONSTANT, constant_id)
                    constant = epanet_api.MSXgetconstant(idx)

                    constant = uncertainty.apply(constant)
                    epanet_api.MSXsetconstant(idx, constant)

            if self._global_parameters is not None:
                self._global_parameters.set_random_generator(np_rand_gen)

                self._cache_msx_links_parameters = {}
                num_params = epanet_api.MSXgetcount(EpanetConstants.MSX_PARAMETER)
                parameters_pipes = [np.array([epanet_api.MSXgetparameter(EpanetConstants.MSX_LINK,
                                                                         pipe_idx,
                                                                         param_idx + 1)
                                              for param_idx in range(num_params)])
                                     for pipe_idx in epanet_api.get_all_pipes_idx()]
                for i, pipe_idx in enumerate(epanet_api.get_all_pipes_idx()):
                    if len(parameters_pipes[i]) == 0:
                        continue

                    self._cache_msx_links_parameters[pipe_idx] = parameters_pipes[i]
                    parameters_pipes_val = self._global_parameters.apply_batch(parameters_pipes[i])
                    for param_idx, param_value in enumerate(parameters_pipes_val):
                        epanet_api.MSXsetparameter(EpanetConstants.MSX_LINK, pipe_idx,
                                                   param_idx + 1, param_value)

                self._cache_msx_tanks_parameters = {}
                num_params = epanet_api.MSXgetcount(EpanetConstants.MSX_PARAMETER)
                parameters_tanks = [np.array([epanet_api.MSXgetparameter(EpanetConstants.MSX_NODE,
                                                                         tank_idx + 1, param_idx + 1)
                                              for param_idx in range(num_params)])
                                     for tank_idx in range(epanet_api.get_num_tanks())]
                for i, tank_idx in enumerate(epanet_api.get_all_tanks_idx()):
                    if parameters_tanks[i] is None or len(parameters_tanks[i]) == 0:
                        continue

                    self._cache_msx_tanks_parameters[tank_idx] = parameters_tanks[i]
                    parameters_tanks_val = self._global_parameters.apply_batch(parameters_tanks[i])
                    for idx, val in enumerate(parameters_tanks_val):
                        epanet_api.MSXsetparameter(EpanetConstants.MSX_NODE, tank_idx, idx + 1, val)

            if self._local_parameters is not None:
                self._local_parameters.set_random_generator(np_rand_gen)
                self._cache_msx_links_parameters = {}
                self._cache_msx_tanks_parameters = {}

                for (param_id, item_type, item_id), uncertainty in self._local_parameters.items():
                    idx, = epanet_api.MSXgetindex(EpanetConstants.MSX_PARAMETER, param_id)

                    if item_type == EpanetConstants.MSX_NODE:
                        item_idx = epanet_api.get_node_idx(item_id)
                    elif item_type == EpanetConstants.MSX_LINK:
                        item_idx = epanet_api.get_link_idx(item_id)
                    else:
                        raise ValueError(f"Unknown item type '{item_type}' must be either " +
                                         "EpanetConstants.MSX_NODE or EpanetConstants.MSX_LINK")

                    parameter = epanet_api.MSXgetparameter(item_type, item_idx, idx)
                    if item_type == EpanetConstants.MSX_NODE:
                        self._cache_msx_tanks_parameters[item_idx] = parameter
                    elif item_type == EpanetConstants.MSX_LINK:
                        self._cache_msx_links_parameters[item_idx] = parameter

                    parameter = uncertainty.apply(parameter)
                    epanet_api.MSXsetparameter(item_type, item_idx, idx, parameter)

            if self._local_msx_patterns is not None:
                self._local_msx_patterns.set_random_generator(np_rand_gen)
                self._cache_msx_patterns = {}

                for pattern_id, uncertainty in self._local_msx_patterns.items():
                    pattern_idx = epanet_api.MSXgetindex(EpanetConstants.MSX_PATTERN, pattern_id)
                    pattern = np.array(epanet_api.get_msx_pattern(pattern_idx))
                    self._cache_msx_patterns[pattern_idx] = np.copy(pattern)

                    pattern = uncertainty.apply_batch(pattern)
                    epanet_api.MSXsetpattern(pattern_idx, pattern.tolist(), len(pattern))
        else:
            if self._local_msx_patterns is not None or self._local_parameters is not None or \
                    self._local_constants is not None or self._global_constants is not None or \
                    self._global_parameters is not None:
                warnings.warn("Ignoring EPANET-MSX uncertainties because not .msx file was loaded")



"""
Module provides a base class for events.
"""
from abc import ABC
import math


class Event(ABC):
    """
    Base class for an event.

    Parameters
    ----------
    start_time : `int`
        Starting time (seconds since the simulation start) of this event.
    end_time : `int`, optional
        Time (seconds since the simulation start) when this event ends -- None if it never ends.

        The default is None.
    """
    def __init__(self, start_time: int, end_time: int = None, **kwds):
        if not isinstance(start_time, int) or start_time < 0:
            raise ValueError("'start_time' must be a positive integer specifying the time " +
                             "at which this event starts.")
        if end_time is not None and not isinstance(end_time, int):
            raise ValueError("'end_time' must be either None or a positive integer specifiying " +
                             "the time at which this event ends.")
        if end_time is not None:
            if start_time >= end_time:
                raise ValueError("'start_time' must be smaller than 'end_time'")

        self.__start_time = start_time
        self.__end_time = end_time if end_time is not None else math.inf

        super().__init__(**kwds)

    @property
    def start_time(self) -> int:
        """
        Gets the start time (seconds since the simulation start) of this event.

        Returns
        -------
        `int`
            Start time of this event.
        """
        return self.__start_time

    @property
    def end_time(self) -> int:
        """
        Gets the end time (seconds since the simulation start) of this event.
        float("inf") if it never ends.

        Returns
        -------
        `int`
            End time of this event.
        """
        return self.__end_time

    def get_attributes(self) -> dict:
        """
        Gets all attributes to be serialized -- these attributes are passed to the
        constructor when the object is deserialized.

        Returns
        -------
        `dict`
            Dictionary of attributes -- i.e. pairs of attribute name and value.
        """
        return {"start_time": self.__start_time, "end_time": self.__end_time}

    def __str__(self) -> str:
        return f"start_time: {self.__start_time} end_time: {self.__end_time}"

    def __eq__(self, other) -> bool:
        if not isinstance(other, Event):
            raise TypeError(f"Can not compare 'Event' instance with '{type(other)}' instance")

        return self.__start_time == other.start_time and self.__end_time == other.end_time
